
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#ifndef _MOOF_CORE_HH_
#define _MOOF_CORE_HH_

#include <boost/shared_ptr.hpp>

#include <Moof/Dispatch.hh>
#include <Moof/Error.hh>
#include <Moof/Layer.hh>


namespace Mf {


/**
 * The core is essentially a stack of layers.  While running, it updates each
 * layer from the bottom up every timestep.  It also draws each layer from the
 * bottom up, adhering to the maximum frame-rate.  Events are sent to each layer
 * from the top down until a layer signals the event was handled.  The core is
 * also responsible for firing timers on time.  The core will continue running
 * as long as there are layers on the stack.
 */

class Core
{
public:

	Core();
	
	// loads settings: rngseed, timestep, framerate, showfps
	void init();

	int getFps() const;

	void push(LayerP layer);	// push a layer onto the top
	LayerP pop();				// pop the top layer
	LayerP pop(Layer* layer);	// pops a specific layer and all layers above it
	void clear();				// remove all layers (the core will stop)

	int getSize() const;		// get the size of the stack

	// set this machine in motion
	void run();

	Dispatch::Handler addHandler(const std::string& event,
			const Dispatch::Function& callback);
	Dispatch::Handler addHandler(const std::string& event,
			const Dispatch::Function& callback, Dispatch::Handler handler);

	void dispatch(const std::string& event,
			const Dispatch::Message* message = 0);

private:

	class Impl;
	boost::shared_ptr<Impl> mImpl;
};

extern Core core;


/*
 * Some classes and subsystems require certain backend libraries to be
 * initialized.  This is the mechanism to accomplish that.  Classes which rely
 * on any backend libraries just need to instantiate this class as a member.
 * Backend cleanup will occur automagically when there are no more instances.
 */

class Backend
{
public:

	Backend();
	~Backend();

	static bool check(Error& error);
};


} // namespace Mf

#endif // _MOOF_CORE_HH_

/** vim: set ts=4 sw=4 tw=80: *************************************************/

